{-
Author:  Christian Sternagel <c.sternagel@gmail.com> (2009-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2009-2014)
License: LGPL (see file COPYING.LESSER)
-}
module Main(main) where 

import Ceta -- the certifier
import System.Environment -- for getArgs
import System.IO -- for file reading
import System.Exit -- for error codes

main = do args <- getArgs
          case Prelude.length args of
            1 -> do problemString <- readFile (args !! 0)
                    start False problemString
            2 -> do let flag = args !! 0
                    if flag == "--allow-assumptions"
                      then do problemString <- readFile (args !! 1)
                              start True problemString
                      else do hPutStrLn stderr ("flag '" ++ flag ++ "' not supported")
                              exitWith (ExitFailure 4)
            _ -> do hPutStrLn stderr "usage: ceta [--allow-assumptions] [--safety] problemFile.xml     (CeTA version 2.28)"
                    exitWith (ExitFailure 4)

start a problemString = 
    do case certify_proof a problemString of
         Sumbot (Inr (Certified prf)) -> 
             do putStrLn ("CERTIFIED " ++ prf)
                exitWith ExitSuccess
         Sumbot (Inr (Error message)) -> 
             do putStrLn "REJECTED"
                hPutStrLn stderr message
                exitWith (ExitFailure 1)
         Sumbot (Inr (Unsupported message)) -> 
             do putStrLn "UNSUPPORTED" 
                hPutStrLn stderr message
                exitWith (ExitFailure 2)
         Sumbot (Inl message) -> 
             do putStrLn "UNKNOWN ERROR"
                hPutStrLn stderr message
                exitWith (ExitFailure 3)


