theory LTS
imports 
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  "$AFP/Show/Show_Instances"
  "$AFP/Derive/Comparator_Generator/Compare_Generator"
  Formula
  "../Orderings/Non_Inf_Order"
begin

datatype 'v trans_var = Pre 'v | Post 'v | Intermediate 'v
derive "show" trans_var
derive compare_order trans_var

type_synonym ('f,'v,'t) transition_formula = "('f,'v trans_var,'t) formula"

datatype ('f,'v,'t,'l) transition = Transition
 (source: 'l)
 (target: 'l)
 (formula: "('f,'v trans_var,'t) formula")

(* derive "show" "('f,'v,'t,'l) transition"
*)
datatype ('v,'t,'d,'l) state = State (valuation: "('v,'t,'d) valuation") (location: 'l)

abbreviation relabel_state
  where "relabel_state f s \<equiv> State (valuation s) (f (location s))"

lemma valuation_relabel_state[simp]: "valuation (relabel_state f s) = valuation s"
  and location_relabel_state[simp]: "location (relabel_state f s) = f (location s)"
  by auto

record ('f,'v,'t,'l) lts =
  initial :: 'l
  transitions :: "('f,'v,'t,'l) transition set"

definition nodes_lts :: "('f,'v,'t,'l) lts \<Rightarrow> 'l set" where
  "nodes_lts P = (source ` transitions P)"

definition subset_lts :: "('f,'v,'t,'l) lts \<Rightarrow> ('f,'v,'t,'l) lts \<Rightarrow> bool" where
  "subset_lts P Q = (initial P = initial Q \<and> transitions P \<subseteq> transitions Q)"

lemma subset_ltsD: assumes "subset_lts P Q" shows "initial P = initial Q" "transitions P \<subseteq> transitions Q"
  using assms unfolding subset_lts_def by auto

definition delete_transitions :: "('f,'v,'t,'l) lts \<Rightarrow> ('f,'v,'t,'l) transition set \<Rightarrow> ('f,'v,'t,'l) lts" where
  "delete_transitions lts ts = \<lparr>lts.initial = lts.initial lts, 
    transitions = transitions lts - ts\<rparr>"

lemma delete_transitions_subset: "subset_lts (delete_transitions lts ts) lts"
  unfolding subset_lts_def delete_transitions_def by auto

lemma delete_transistions_eq[simp]:
  "delete_transitions P (transitions P \<inter> TD) = delete_transitions P TD"
   unfolding subset_lts_def delete_transitions_def by auto

context pretheory begin

definition state :: "('v,'t,'d,'l) state \<Rightarrow> bool" where [simp]: "state s \<equiv> assignment (valuation s)"

definition lts :: "('f,'v,'t,'l) lts \<Rightarrow> bool" where
  "lts P \<equiv> \<forall> \<tau> \<in> transitions P. formula (transition.formula \<tau>)"

fun pre_post_inter where
  "pre_post_inter \<alpha> \<beta> \<gamma> (Pre x,t) = \<alpha> (x,t)"
| "pre_post_inter \<alpha> \<beta> \<gamma> (Post x,t) = \<beta> (x,t)"
| "pre_post_inter \<alpha> \<beta> \<gamma> (Intermediate x,t) = \<gamma> (x,t)"

context
  fixes \<alpha> \<beta> \<gamma> :: "('v,'t,'d) valuation"
begin

abbreviation "\<delta> \<equiv> pre_post_inter \<alpha> \<beta> \<gamma>"

lemma pre_post_inter_update[simp]:
  "\<delta>((Pre x,ty):=a) = pre_post_inter (\<alpha>((x,ty):=a)) \<beta> \<gamma>" (is "?l1 = ?r1")
  "\<delta>((Post x,ty):=a) = pre_post_inter \<alpha> (\<beta>((x,ty):=a)) \<gamma>" (is "?l2 = ?r2")
  "\<delta>((Intermediate x,ty):=a) = pre_post_inter \<alpha> \<beta> (\<gamma>((x,ty):=a))" (is "?l3 = ?r3")
proof-
  have "?l1 xx = ?r1 xx \<and> ?l2 xx = ?r2 xx \<and> ?l3 xx = ?r3 xx" for xx by(cases xx, cases "fst xx", auto)
  then show "?l1 = ?r1" "?l2 = ?r2" "?l3 = ?r3" by auto
qed

lemma assignment_pre_post_inter[simp]:
  "assignment \<delta> \<longleftrightarrow> assignment \<alpha> \<and> assignment \<beta> \<and> assignment \<gamma>" (is "?l \<longleftrightarrow> ?r")
proof(intro iffI)
  assume ?r
  then have "\<delta> (x, t) \<in> type_values t" for x t by(cases x,auto)
  then show ?l by auto
next
  assume ?l
  note this[unfolded assignment_def, rule_format]
  from this[of "Pre _"] this[of "Post _"] this[of "Intermediate _"] show ?r by auto
qed

lemma eval_pre_post_inter_rename_vars[simp]:
  shows "eval \<delta> (rename_vars Pre e) = eval \<alpha> e" (is ?a)
    and "eval \<delta> (rename_vars Post e) = eval \<beta> e" (is ?b)
    and "eval \<delta> (rename_vars Intermediate e) = eval \<gamma> e" (is ?c)
proof -
  have "?a \<and> ?b \<and> ?c"
  proof(induct e)
  case (Var xx)
    show ?case by (cases xx, auto simp: rename_vars_exp_def)
  case IH: (Fun f es)
    from IH have 1: "map (eval \<delta>) (map (rename_vars Pre) es) = map (eval \<alpha>) es" by(induct es; auto)
    from IH have 2: "map (eval \<delta>) (map (rename_vars Post) es) = map (eval \<beta>) es" by(induct es; auto)
    from IH have 3: "map (eval \<delta>) (map (rename_vars Intermediate) es) = map (eval \<gamma>) es" by(induct es; auto)
    show ?case unfolding rename_vars_Fun eval.simps 1 2 3 by auto
  qed
  then show ?a ?b ?c by auto
qed

lemma pre_post_inter_satisfies_rename_vars_pre[simp]:
  "\<delta> \<Turnstile> rename_vars Pre \<phi> \<longleftrightarrow> \<alpha> \<Turnstile> \<phi>" (is "?l \<longleftrightarrow> ?r")
proof (intro iffI satisfiesI)
  assume l: ?l
  fix cl assume "cl \<in> set \<phi>"
  then have "map (rename_vars Pre) cl \<in> set (rename_vars Pre \<phi>)" by(auto simp: rename_vars_formula_def)
  from satisfiesD[OF l this]
  have "\<exists>lit\<in>set (map (rename_vars Pre) cl). eval_true (pre_post_inter \<alpha> \<beta> \<gamma>) lit" by auto
  then obtain lit
  where lit: "lit \<in> set cl" and "eval_true \<alpha> lit" by auto
  then show "\<exists>lit\<in>set cl. eval_true \<alpha> lit" by auto
next
  assume r: ?r
  fix cl assume "cl \<in> set (rename_vars Pre \<phi>)"
  then obtain cl'
  where 1: "cl' \<in> set \<phi>"
    and 2: "cl = map (rename_vars Pre) cl'" by (auto simp: rename_vars_formula_def)
  from satisfiesD[OF r 1] obtain lit'
  where 3: "lit' \<in> set cl'"
    and 4: "eval_true \<alpha> lit'" by auto
  from 2 3 have 5: "rename_vars Pre lit' \<in> set cl" by auto
  from 4 show "\<exists>lit\<in>set cl. eval_true (pre_post_inter \<alpha> \<beta> \<gamma>) lit"
  by (intro bexI[OF _ 5], auto)
qed
lemma pre_post_inter_satisfies_rename_vars_post[simp]:
  "\<delta> \<Turnstile> rename_vars Post \<phi> \<longleftrightarrow> \<beta> \<Turnstile> \<phi>" (is "?l \<longleftrightarrow> ?r")
proof (intro iffI satisfiesI)
  assume l: ?l
  fix cl assume "cl \<in> set \<phi>"
  then have "map (rename_vars Post) cl \<in> set (rename_vars Post \<phi>)" by(auto simp: rename_vars_formula_def)
  from satisfiesD[OF l this]
  have "\<exists>lit\<in>set (map (rename_vars Post) cl). eval_true (pre_post_inter \<alpha> \<beta> \<gamma>) lit" by auto
  then obtain lit
  where lit: "lit \<in> set cl" and "eval_true \<beta> lit" by auto
  then show "\<exists>lit\<in>set cl. eval_true \<beta> lit" by auto
next
  assume r: ?r
  fix cl assume "cl \<in> set (rename_vars Post \<phi>)"
  then obtain cl'
  where 1: "cl' \<in> set \<phi>"
    and 2: "cl = map (rename_vars Post) cl'" by (auto simp: rename_vars_formula_def)
  from satisfiesD[OF r 1] obtain lit'
  where 3: "lit' \<in> set cl'"
    and 4: "eval_true \<beta> lit'" by auto
  from 2 3 have 5: "rename_vars Post lit' \<in> set cl" by auto
  from 4 show "\<exists>lit\<in>set cl. eval_true (pre_post_inter \<alpha> \<beta> \<gamma>) lit"
  by (intro bexI[OF _ 5], auto)
qed

end

definition transition_step where
  "transition_step \<tau> \<equiv> { (s,t). \<exists> \<gamma>.
   state s \<and> state t \<and> assignment \<gamma> \<and>
   source \<tau> = location s \<and> target \<tau> = location t \<and>
   pre_post_inter (valuation s) (valuation t) \<gamma> \<Turnstile> transition.formula \<tau> }"

lemma mem_transition_stepI[intro]:
   "state s \<Longrightarrow> state t \<Longrightarrow> assignment \<gamma> \<Longrightarrow>
    source \<tau> = location s \<Longrightarrow>
    target \<tau> = location t \<Longrightarrow>
    pre_post_inter (valuation s) (valuation t) \<gamma> \<Turnstile> transition.formula \<tau> \<Longrightarrow>
    (s,t) \<in> transition_step \<tau>"
  unfolding transition_step_def by auto

lemma mem_transition_stepD[dest]:
  assumes "(s,t) \<in> transition_step \<tau>"
  shows "state s" "state t"
    "location s = source \<tau>" "location t = target \<tau>"
    "\<exists>\<gamma>. assignment \<gamma> \<and> pre_post_inter (valuation s) (valuation t) \<gamma> \<Turnstile> transition.formula \<tau>"
  using assms[unfolded transition_step_def] by auto

lemma mem_transition_stepE[elim]:
  assumes base: "(s,t) \<in> transition_step \<tau>"
      and step: "\<And>\<gamma>. state s \<Longrightarrow> state t \<Longrightarrow> assignment \<gamma> \<Longrightarrow>
    location s = source \<tau> \<Longrightarrow> location t = target \<tau> \<Longrightarrow>
    pre_post_inter (valuation s) (valuation t) \<gamma> \<Turnstile> transition.formula \<tau> \<Longrightarrow> P"
  shows P
  using base[unfolded transition_step_def] step by auto

lemma transition_step_dom: "transition_step \<tau> \<subseteq> Collect state \<times> Collect state" by auto

abbreviation transitions_step where "transitions_step \<tau>s \<equiv> UNION \<tau>s transition_step"
abbreviation transitions_steps where "transitions_steps \<tau>s \<equiv> (transitions_step \<tau>s)\<^sup>*"

lemma transitions_step_dom: "transitions_step \<tau>s \<subseteq> Collect state \<times> Collect state" by auto

definition transition :: "('f,'v,'t,'l) lts \<Rightarrow> ('v,'t,'d,'l) state rel" where
  "transition P \<equiv> transitions_step (transitions P)"

lemma mem_transitionI[intro!]:
  "\<tau> \<in> transitions P \<Longrightarrow> (l,r) \<in> transition_step \<tau> \<Longrightarrow> (l,r) \<in> transition P"
and mem_transitionE[elim!]:
  "(l,r) \<in> transition P \<Longrightarrow> (\<And>\<tau>. \<tau> \<in> transitions P \<Longrightarrow> (l,r) \<in> transition_step \<tau> \<Longrightarrow> foo) \<Longrightarrow> foo"
  unfolding transition_def by auto

lemma mem_transition_location:
  "(s,t) \<in> transition P \<Longrightarrow> location s \<in> nodes_lts P" by (auto simp: nodes_lts_def)

lemma mem_transition_dom:
  assumes "(s,t) \<in> transition P"
  shows "state s" and "state t"
  using assms by auto

lemma transition_dom: "transition P \<subseteq> Collect state \<times> Collect state" by auto

definition initial_states :: "('f,'v,'t,'l) lts \<Rightarrow> ('v,'t,'d,'l) state set" where
  "initial_states P = { State \<alpha> (initial P) | \<alpha>. assignment \<alpha>}"

lemma initial_states: "initial_states P \<subseteq> Collect state" unfolding initial_states_def by auto

abbreviation lts_termination :: "('f,'v,'t,'l) lts \<Rightarrow> bool" where
  "lts_termination P \<equiv> SN_on (transition P) (initial_states P)"

abbreviation reachable_states :: "('f,'v,'t,'l) lts \<Rightarrow> ('v,'t,'d,'l) state set" where
  "reachable_states P \<equiv> (transition P)\<^sup>* `` initial_states P"

lemma reachable_state: assumes "s \<in> reachable_states P" shows "state s"
proof -
  from assms obtain t where steps: "(t, s) \<in> (transition P)^*"
    and t: "assignment (valuation t)" unfolding initial_states_def by auto
  from steps show ?thesis
  proof (induct rule: rtrancl_induct)
    case (step s u)
    from step(2)
    show ?case unfolding transition_def transition_step_def Let_def by auto
  qed (insert t, auto)
qed

definition disj_invariant_set :: "('f,'v,'t,'l) lts \<Rightarrow> ('l \<Rightarrow> ('f,'v,'t) formula set) \<Rightarrow> bool" where
  "disj_invariant_set P I \<equiv> \<forall> \<alpha> l. (State \<alpha> l) \<in> reachable_states P \<longrightarrow> (\<exists> \<phi> \<in> I l. \<alpha> \<Turnstile> \<phi>)"

definition disj_invariant :: "('f,'v,'t,'l) lts \<Rightarrow> ('l \<Rightarrow> ('f,'v,'t) formula list) \<Rightarrow> bool" where
  "disj_invariant P I \<equiv> \<forall> \<alpha> l. (State \<alpha> l) \<in> reachable_states P \<longrightarrow> (\<exists> \<phi> \<in> set (I l). \<alpha> \<Turnstile> \<phi>)"

lemma disj_invariant_is_disj_invariant_set[simp]:
"disj_invariant_set P (set \<circ> I) = disj_invariant P I"
unfolding disj_invariant_set_def disj_invariant_def by auto

lemma disj_invariantD: "disj_invariant P I \<Longrightarrow> State \<alpha> l \<in> reachable_states P \<Longrightarrow> \<exists> \<phi> \<in> set (I l). \<alpha> \<Turnstile> \<phi>"
  unfolding disj_invariant_def by auto
lemma disj_invariant_setD: "disj_invariant_set P I \<Longrightarrow> State \<alpha> l \<in> reachable_states P \<Longrightarrow> \<exists> \<phi> \<in> I l. \<alpha> \<Turnstile> \<phi>"
  unfolding disj_invariant_set_def by auto

lemma reachable_states_declines_by_subset:
  assumes "transitions P \<subseteq> transitions Q"
      and "initial_states P \<subseteq> initial_states Q"
  shows "reachable_states P \<subseteq> reachable_states Q"
proof
  have tr:"transition P \<subseteq> transition Q" using assms(1) by auto
  fix x
  assume "x \<in> reachable_states P"
  then obtain start where "(start,x) \<in> (transition P)\<^sup>*" "start \<in> initial_states P" by auto
  hence "(start,x) \<in> (transition Q)\<^sup>*" "start \<in> initial_states Q"
    using rtrancl_mono[OF tr] assms(2) by auto
  thus "x \<in> reachable_states Q" by auto
qed

end

subsection {* An Implementation of Labeled Transition Systems *}

type_synonym ('f,'v,'t,'l,'tr) transitions_impl = "('tr \<times> ('f,'v,'t,'l) transition) list"

datatype ('f,'v,'t,'l,'tr) lts_impl = Lts_Impl
  (initial: 'l)
  (transitions_impl: "('f,'v,'t,'l,'tr) transitions_impl")

definition "transition_of Pi \<equiv> the \<circ> map_of (transitions_impl Pi)"

fun lts_of :: "('f,'v,'t,'l,'tr) lts_impl \<Rightarrow> ('f,'v,'t,'l) lts" where
  "lts_of Pi = \<lparr>
    lts.initial = lts_impl.initial Pi,
    transitions = { \<tau>. \<tau> \<in> snd ` set (transitions_impl Pi) } \<rparr>"

fun succ_transitions :: "('f,'v,'t,'l,'tr) lts_impl \<Rightarrow> 'l \<Rightarrow> ('f,'v,'t,'l) transition list" where
  "succ_transitions (Lts_Impl i ts) l = [ \<tau>. (tr,\<tau>) \<leftarrow> ts, source \<tau> = l ]"

lemma succ_transitions[simp]:
  "set (succ_transitions Pi l) =
   { \<tau>. \<tau> \<in> transitions (lts_of Pi) \<and> source \<tau> = l}" by (cases Pi, auto)

type_synonym ('f,'v,'t,'l) disjInv = "('l \<Rightarrow> ('f,'v,'t) exp list list list)"
context pretheory begin

(* TODO: a nicer checking function *)
definition "lts_impl Pi \<equiv> lts (lts_of Pi)"

definition "lts_impl_checker Pi = check (lts_impl Pi) (shows ''ill-formed transition system'')"

lemma lts_impl_checker[intro]: "isOK (lts_impl_checker Pi) \<Longrightarrow> lts_impl Pi"
  by(simp add: lts_impl_checker_def)

lemma lts_impl[intro]: "lts_impl Pi \<Longrightarrow> lts (lts_of Pi)" 
  unfolding lts_impl_def by auto

definition "disj_invariant_checker f \<equiv> 
   \<forall> P. lts (lts_of P) \<longrightarrow> (case f P of Inl _ \<Rightarrow> True
  | Inr I \<Rightarrow> disj_invariant (lts_of P) I \<and> (\<forall> l. list_all formula (I l)))"


subsection {* Refining transition formulas *}

definition "refine_transition_formulas P f \<equiv>
   Lts_Impl (lts_impl.initial P)
    (map (\<lambda>(tr,\<tau>). (tr, Transition (source \<tau>) (target \<tau>) (transition.formula \<tau> \<and>\<^sub>f f tr)))
    (transitions_impl P))"

lemma initial_refine_transition_formulas[simp]:
  "lts_impl.initial (refine_transition_formulas P f) = lts_impl.initial P"
  by (auto simp: refine_transition_formulas_def)

end


end
