(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2016)
License: LGPL (see file COPYING.LESSER)
*)

theory Gtcap_Impl
imports
  Gtcap
  "../Rewriting/Tcap_Impl"
  "$AFP/Transitive-Closure/Transitive_Closure_List_Impl"
  "../Rewriting/Term_Impl"
begin

definition "gt1 = map (\<lambda>r. case r of (s, t) \<Rightarrow> (root s, root t))"

lemma gt1_iff_GT1 [simp]:
  "set (gt1 rs) = GT1 (set rs)"
by (auto simp: GT1_def gt1_def)

definition "mk_gt_fun rs =
  (let in_trancl = memo_list_trancl (gt1 rs) in
  (\<lambda>f g. g \<in> set (in_trancl f)))"

lemma mk_gt_fun_iff [simp]:
  "mk_gt_fun rs f g \<longleftrightarrow> (f, g) \<in> (GT1 (set rs))\<^sup>+"
by (auto simp: mk_gt_fun_def memo_list_trancl)

lemma
  "(s, t) \<in> GT (set rs) \<longleftrightarrow>
  is_Var s \<or> is_Var t \<or> mk_gt_fun rs (root s) (root t) \<or> mk_gt_fun rs (root s) None
  \<or> mk_gt_fun rs None (root t) \<or> mk_gt_fun rs None None"
by (auto simp: GT_def)

fun gt_term
where
  "gt_term gt_fun rm (Fun f ss) (Fun g ts) =
    (let fm = (f, length ss) in
    let root1 = Some fm; root2 = Some (g, length ts) in
    (gt_fun root1 root2 \<or> gt_fun root1 None) \<and>
    (let h = GCFun f (map (tcapRM2 rm) ss) in
    \<exists>r\<in>set (rm fm). Ground_Context.match h (fst r)))"
| "gt_term gt_fun rm _ _ = True"

fun rd_impl
where
  "rd_impl gt (Fun f ss, Fun g ts) =
    (if f = g \<and> length ss = length ts \<and> \<not> gt (Fun f ss) (Fun g ts) then concat (map (rd_impl gt) (zip ss ts))
    else [(Fun f ss, Fun g ts)])"
| "rd_impl _ (s, t) = [(s, t)]"

lemma rd_impl:
  fixes R :: "('f, 'v) trs"
  assumes gt: "\<And>s t. gt s t \<longleftrightarrow>
    (s, t) \<in> GT R \<inter> {(s, t :: ('f, 'v) term). \<exists>(l, r) \<in> R. Ground_Context.match (tcap R s) l}"
    (is "\<And>s t. gt s t \<longleftrightarrow> (s, t) \<in> ?S")
    and "(s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
  shows "\<forall>(u, v) \<in> set (rd_impl gt (s, t)). (u \<cdot> \<sigma>, v \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
proof -
  interpret reachability_decomposition R "?S"
    by (intro reachability_decomposition_Int rd_root rd_tcap)
  have "\<And>s t. rd_impl gt (s, t) = reach_decomp s t"
  proof -
    fix s t
    show "?thesis s t"
    unfolding reach_decomp_reach_decomp'_conv
    apply (induct ("(s, t)") rule: reach_decomp'.induct)
    apply (case_tac "(Fun f ss, Fun g ts) \<in> ?S")
    apply (auto simp only: reach_decomp'.simps rd_impl.simps gt [symmetric])
    apply (auto intro!: arg_cong [of _ _ concat])
    done
  qed
  then show ?thesis using assms by (auto dest: reach_decomp_sound)
qed

end
