(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2013-2016)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2012-2015)
Author:  Thomas Sternagel <thomas.sternagel@uibk.ac.at> (2012)
License: LGPL (see file COPYING.LESSER)
*)
theory Check_CRP
imports 
  Check_Termination
begin


subsection \<open>Proving confluence of TRSs\<close>

datatype ('f, 'l, 'v) cr_proof =
    SN_WCR "('f, 'l) lab join_info" "('f, 'l, 'v) trs_termination_proof" 
  | Weakly_Orthogonal
  | Strongly_Closed "nat"
  | Rule_Labeling "(('f, 'l) lab, 'v) rule_lab_repr" "(('f, 'l) lab ,string) crit_pair_joins" "('f, 'l, 'v) trs_termination_proof option"
  | Rule_Labeling_Conv "(('f, 'l) lab, 'v) rule_lab_repr" "(('f, 'l) lab ,string) crit_peak_convs" "(nat \<times> ('f, 'l, 'v) trs_termination_proof) option"
  | Redundant_Rules "('f, 'l, 'v) trsLL" "nat" "(('f, 'l) lab, 'v) eseq list" "('f, 'l, 'v) cr_proof"
  | Parallel_Closed "nat option"

primrec
  check_cr_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,key,countable}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> (('f,label_type)lab,string) rules \<Rightarrow>
     ('f, label_type, string) cr_proof \<Rightarrow>
     shows check" 
where
  "check_cr_proof a i I J R (SN_WCR joins_i prf) = debug (i []) ''SN_WCR'' (do {
         let tp = tp_ops.mk I False [] R [];
         check_trs_termination_proof I J a (i +@+ shows_string ''.1'') tp prf
           <+? (\<lambda>s. i +@+ '': error below strong normalization + wcr'' +#+ shows_nl
             +@+ indent s);
         check_critical_pairs R (critical_pairs_impl R R) joins_i
           <+? (\<lambda>s. i +@+ shows ''error when proving local confluence of '' +@+ shows_tp I tp +@+ shows_nl +@+ indent s)
   })" 
| "check_cr_proof a i I J R Weakly_Orthogonal = debug (i []) ''Weakly Orthogonal'' ( do {
           check_weakly_orthogonal R 
         }  <+? (\<lambda>s. i +@+ shows '': error in checking weakly orthogonality of the '' +@+ shows_trs R
             +@+ indent s))"
| "check_cr_proof a i I J R (Strongly_Closed n) = debug (i []) ''Strongly Closed'' ( do {
           check_strongly_closed R n
         }  <+? (\<lambda>s. i +@+ shows '': error in checking strong closedness for the '' +@+ shows_trs R
             +@+ indent s))"
| "check_cr_proof a i I J R (Rule_Labeling rl joins prf) = debug (i []) ''Rule Labeling'' ( do {
           (case prf of
             None \<Rightarrow> check_linear_trs R
           | Some prf \<Rightarrow> do {
             check_left_linear_trs R;
             let (Rnd, Rd) = List.partition (\<lambda>lr. linear_term (snd lr)) R;
             let tp = tp_ops.mk I False [] Rd Rnd;
             check_trs_termination_proof I J a (i +@+ shows_string ''.1'') tp prf
               <+? (\<lambda>s. i +@+ '': error below relative termination for rule labeling'' +#+ shows_nl +@+ indent s)
             }
           );
           check_rule_labeling_eld R rl joins
             <+? (\<lambda>s. i +@+ shows '': error in checking decreasingness of CPs using rule labeling for the ''
             +@+ shows_trs R +@+ indent s)
           })"
| "check_cr_proof a i I J R (Rule_Labeling_Conv rl convs nprf) = debug (i []) ''Rule Labeling'' ( do {
           (case nprf of
             None \<Rightarrow> do {
             check_linear_trs R;
             check_rule_labeling_eldc R rl convs None
               <+? (\<lambda>s. i +@+ shows '': error in checking decreasingness of CPs using rule labeling for the ''
               +@+ shows_trs R +@+ indent s)
             }
           | Some (n, prf) \<Rightarrow> do {
             check_left_linear_trs R;
             let (Rnd, Rd) = List.partition (\<lambda>lr. linear_term (snd lr)) R;
             let tp = tp_ops.mk I False [] Rd Rnd;
             check_trs_termination_proof I J a (i +@+ shows_string ''.1'') tp prf
               <+? (\<lambda>s. i +@+ '': error below relative termination for rule labeling'' +#+ shows_nl +@+ indent s);
             check_rule_labeling_eldc R rl convs (Some n)
               <+? (\<lambda>s. i +@+ shows '': error in checking decreasingness of CPs using rule labeling for the ''
               +@+ shows_trs R +@+ indent s)
             }
           )})"
| "check_cr_proof a i I J R (Redundant_Rules RS n convs prf) = debug (i []) ''Redundant Rules'' (do {
         check_cr_proof a (i +@+ shows_string ''.1'') I J RS prf
           <+? (\<lambda>s. i +@+ shows '': error below confluence of modified TRS'' +@+ shows_nl +@+ shows_trs RS +@+ indent s);
         check_redundant_rules R RS n convs
           <+? (\<lambda>s. i +@+ shows '': error in checking redundant rules transformation of the '' 
                  +@+ shows_trs R +@+ shows ''transformed to the '' +@+ shows_trs RS +@+ indent s)
         })"
| "check_cr_proof a i I J R (Parallel_Closed n) = debug (i []) ''Parallel Closed'' ( do {
           check_parallel_closed R n
         }  <+? (\<lambda>s. i +@+ shows '': error in checking parallel closedness for the '' +@+ shows_trs R
             +@+ indent s))"

fun cr_assms :: "bool \<Rightarrow> ('f, 'l, 'v) cr_proof \<Rightarrow> (('f, 'l, 'v) assm) list" where
  "cr_assms a (SN_WCR _ p) = sn_assms a p"
| "cr_assms a Weakly_Orthogonal = []"
| "cr_assms a (Strongly_Closed _) = []"
| "cr_assms a (Rule_Labeling _ _ (Some p)) = sn_assms a p"
| "cr_assms a (Rule_Labeling _ _ None) = []"
| "cr_assms a (Rule_Labeling_Conv _ _ (Some (n, p))) = sn_assms a p"
| "cr_assms a (Rule_Labeling_Conv _ _ None) = []"
| "cr_assms a (Redundant_Rules _ _ _ p) = cr_assms a p"
| "cr_assms a (Parallel_Closed _) = []"

lemma check_cr_proof_with_assms_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (cr_assms a prf). holds p" (is "?P a prf")
    and ok: "isOK (check_cr_proof a i I J R prf)"
  shows "CR (rstep (set R))"
proof -
  interpret tp_spec I by fact
  from ok fin show ?thesis
  proof (induct "prf" arbitrary: i R)
    case (SN_WCR joins_i prof i)
    from SN_WCR(1)
    have cp: "isOK(check_critical_pairs R (critical_pairs_impl R R) joins_i)"
      and ok: "isOK(check_trs_termination_proof I J a (i +@+ shows_string ''.1'') (mk False [] R []) prof)" by (auto simp: Let_def)
    from SN_WCR(2) have a: "\<forall> a \<in> set (sn_assms a prof). holds a" by auto
    from check_critical_pairs[OF cp] have WCR: "WCR (rstep (set R))" .
    from check_trs_termination_proof_with_assms[OF I J ok a] have SN: "SN (rstep (set R))" by simp
    from Newman[OF SN WCR] show ?case .
  next
    case Weakly_Orthogonal
    hence "isOK(check_weakly_orthogonal R)" by auto
    from check_weakly_orthogonal[OF this] show ?case by simp
  next
    case (Strongly_Closed n)
    hence "isOK(check_strongly_closed R n)" by auto
    from check_strongly_closed[OF this] show ?case .
  next
    case (Rule_Labeling rl joins p)
    let ?RndRd = "List.partition (\<lambda>lr. linear_term (snd lr)) R"
    let ?Rnd = "fst ?RndRd"
    let ?Rd = "snd ?RndRd"
    from Rule_Labeling(1) have dec:"isOK(check_rule_labeling_eld R rl joins)" by (auto simp: Let_def)
    have "SN_rel (rstep (R_d (set R))) (rstep (R_nd (set R))) \<and> left_linear_trs (set R)"
    proof (cases p)
      case (Some prof)
      with Rule_Labeling(1) have
        sn:"isOK(check_trs_termination_proof I J a (i +@+ shows_string ''.1'') (mk False [] ?Rd ?Rnd) prof)"
        and ll: "left_linear_trs (set R)"
        by (auto simp: Let_def)
      from Rule_Labeling(2) Some have a:"\<forall> a \<in> set (sn_assms a prof). holds a" by auto
      have "set ?Rd = R_d (set R)" and "set ?Rnd = R_nd (set R)"
        unfolding R_d_def R_nd_def by auto
      with check_trs_termination_proof_with_assms[OF I J sn a] ll
      show ?thesis by auto
    next
      case None
      with Rule_Labeling(1) have "linear_trs (set R)" by auto
      with linear_imp_SN_rel_d_nd show ?thesis
        unfolding linear_trs_def left_linear_trs_def by auto
    qed
    with check_rule_labeling_eld[OF dec] show ?case by auto
  next
    case (Rule_Labeling_Conv rl convs p)
    let ?RndRd = "List.partition (\<lambda>lr. linear_term (snd lr)) R"
    let ?Rnd = "fst ?RndRd"
    let ?Rd = "snd ?RndRd"
    show ?case
    proof (cases p)
      case None
      with Rule_Labeling_Conv(1) have "isOK(check_rule_labeling_eldc R rl convs None)" 
        and "linear_trs (set R)" 
        by (auto simp: Let_def)
      with check_rule_labeling_eldc linear_imp_SN_rel_d_nd show ?thesis 
        unfolding linear_trs_def left_linear_trs_def by blast
    next
      case (Some p')
      then obtain n  pr where p':"p' = (n, pr)" by (cases p') auto
      with Some Rule_Labeling_Conv(1) have eldc:"isOK(check_rule_labeling_eldc R rl convs (Some n))"
        by (auto simp: Let_def)
      from Some p' Rule_Labeling_Conv(1) have
        sn:"isOK(check_trs_termination_proof I J a (i +@+ shows_string ''.1'') (mk False [] ?Rd ?Rnd) pr)"
        and ll: "left_linear_trs (set R)"
        by (auto simp: Let_def)
      from Rule_Labeling_Conv(2) Some p' have a:"\<forall> a \<in> set (sn_assms a pr). holds a" by auto
      have "set ?Rd = R_d (set R)" and "set ?Rnd = R_nd (set R)"
        unfolding R_d_def R_nd_def by auto
      with check_trs_termination_proof_with_assms[OF I J sn a] 
      have "SN_rel (rstep (R_d (set R))) (rstep (R_nd (set R)))" by auto
      from check_rule_labeling_eldc[OF eldc this ll] show ?thesis by auto
    qed
  next
    case (Redundant_Rules RS n convs prof)
    hence "CR (rstep (set RS))" by auto
    moreover from Redundant_Rules have "isOK(check_redundant_rules R RS n convs)" by auto
    ultimately show ?case using check_redundant_rules by blast
  next
    case (Parallel_Closed n)
    hence "isOK(check_parallel_closed R n)" by auto
    from check_parallel_closed[OF this] show ?case .
  qed
qed

lemma cr_assms_False[simp]: "cr_assms False prf = []"
  proof (induct "prf")
    case (Rule_Labeling rl js p)
    show ?case by (cases p) simp_all
  next
    case (Rule_Labeling_Conv rl js p)
    show ?case by (cases p) auto
qed simp_all

lemma check_cr_proof_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
  and ok: "isOK (check_cr_proof False i I J R prf)"
  shows "CR (rstep (set R))"
  by (rule check_cr_proof_with_assms_sound[OF I J _ ok], simp)

datatype (dead 'f, dead 'l, dead 'v, 'q) ncr_proof =
  SN_NWCR "('f, 'l, 'v) trs_termination_proof" 
| Non_Join
    "(('f, 'l) lab, 'v) term"
    "(('f, 'l) lab, 'v) rseq" "(('f, 'l) lab, 'v) rseq" "(('f, 'l) lab, 'v, 'q) non_join_info"
| NCR_Disj_Subtrs "(('f, 'l)lab, 'v) rules" "('f, 'l, 'v, 'q) ncr_proof"
| NCR_Redundant_Rules "('f, 'l, 'v) trsLL" "nat" "('f, 'l, 'v, 'q) ncr_proof"

primrec
  check_ncr_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,key,countable}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> (('f,label_type)lab,string) rules \<Rightarrow>
     ('f, label_type, string, string) ncr_proof \<Rightarrow>
     shows check" 
where
  "check_ncr_proof a i I J R (SN_NWCR prf) = debug (i []) ''SN_NWCR'' (do {
         let tp = tp_ops.mk I False [] R [];
         check_trs_termination_proof I J a (i +@+ shows_string ''.1'') tp prf
           <+? (\<lambda>s. i +@+ '': error below strong normalization + wcr'' +#+ shows_nl
             +@+ indent s);
         check (\<not> isOK(check_critical_pairs_NF R (critical_pairs_impl R R))) (shows ''all critical pairs are joinable'')
           <+? (\<lambda>s. i +@+ shows ''error when disproving local confluence of '' +@+ shows_tp I tp +@+ shows_nl +@+ indent s)
   })" 
| "check_ncr_proof a i I J R (Non_Join s seq1 seq2 prf) = debug (i []) ''Non_Join'' (do {
         check_non_cr R s seq1 seq2 prf
           <+? (\<lambda>s. i +@+ shows ''error when disproving CR of '' +@+ shows_trs R +@+ shows_nl +@+ indent s)
   })" 
| "check_ncr_proof a i I J R (NCR_Disj_Subtrs R' prf) = debug (i []) ''Modularity'' (do {
      check_modularity_ncr R R'  
          <+? (\<lambda>s. i +@+ shows ''error when applying modularity to switch to '' +@+ shows_trs R' +@+ shows_nl +@+ indent s);
      check_ncr_proof a (i +@+ shows_string ''.1'') I J R' prf 
        <+? (\<lambda> s. i +@+ '': error below the modular decomposition'' +#+ shows_nl +@+ indent s)
    })"
| "check_ncr_proof a i I J R (NCR_Redundant_Rules RS n prf) = debug (i []) ''Redundant Rules'' (do {
         check_ncr_proof a (i +@+ shows_string ''.1'') I J RS prf
           <+? (\<lambda>s. i +@+ shows '': error when proving nonconfluence of modified TRS'' +@+ shows_nl +@+ shows_trs RS +@+ indent s);
         check_redundant_rules_ncr R RS n
           <+? (\<lambda>s. i +@+ shows ''error in checking redundant rules transformation of the TRS'' +@+ shows_nl +@+ shows_trs R +@+ indent s)
    })"

primrec ncr_assms :: "bool \<Rightarrow> ('f, 'l, 'v, 'q) ncr_proof \<Rightarrow> (('f, 'l, 'v) assm) list" where
  "ncr_assms a (SN_NWCR p) = sn_assms a p"
| "ncr_assms a (NCR_Disj_Subtrs _ p) = ncr_assms a p"
| "ncr_assms a (Non_Join _ _ _ _)  = []"
| "ncr_assms a (NCR_Redundant_Rules _ _ p) = ncr_assms a p"

lemma check_ncr_proof_with_assms_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (ncr_assms a prf). holds p"
    and ok: "isOK (check_ncr_proof a i I J R prf)"
  shows "\<not> CR (rstep (set R))"
proof -
  interpret tp_spec I by fact
  from ok fin show ?thesis
  proof (induct "prf" arbitrary: i R)
    case (SN_NWCR prof i)
    from SN_NWCR(2) have a: "\<forall> a \<in> set (sn_assms a prof). holds a" by auto
    from SN_NWCR(1)
    have cp: "\<not> isOK(check_critical_pairs_NF R (critical_pairs_impl R R))"
      and ok: "isOK(check_trs_termination_proof I J a (i +@+ shows_string ''.1'') (mk False [] R []) prof)" by (auto simp: Let_def)
    from check_trs_termination_proof_with_assms[OF I J ok a] have SN: "SN (rstep (set R))" by simp
    from check_critical_pairs_NF_SN[OF SN] cp show ?case by simp
  next
    case (Non_Join s seq1 seq2 prof)
    hence ok: "isOK (check_non_cr R s seq1 seq2 prof)" by simp
    from check_non_cr[OF ok] show ?case .
  next 
    case (NCR_Disj_Subtrs R' prof)
    note IH = this
    from IH(3) have a: "\<forall>p\<in>set (ncr_assms a prof). holds p" by auto
    from IH(2) have ok: "isOK (check_modularity_ncr R R')" by simp
    show ?case
      by (rule check_modularity_ncr[OF ok IH(1)[OF _ a] infinite_lab],
      insert IH(2), auto)
  next
    case (NCR_Redundant_Rules RS n prof)
    hence "\<not> CR (rstep (set RS))" by auto
    moreover from NCR_Redundant_Rules have "isOK(check_redundant_rules_ncr R RS n)" by auto
    ultimately show ?case using check_redundant_rules_ncr by auto
  qed
qed

lemma ncr_assms_False[simp]: "ncr_assms False prf = []"
  by (induct "prf") simp_all

lemma check_ncr_proof_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
  and ok: "isOK (check_ncr_proof False i I J R prf)"
  shows "\<not> CR (rstep (set R))"
  by (rule check_ncr_proof_with_assms_sound[OF I J _ ok], simp)

definition default_grd_fun :: "((string,'l)lab,string)term \<Rightarrow> ((string,'l)lab,string)term \<Rightarrow> ((string,'l)lab,string)subst"
  where "default_grd_fun s t \<equiv> let 
           F = funs_rule_list (s,t);
           m = fold (\<lambda> f m. case f of Sharp (UnLab g) \<Rightarrow> max (length g) m | _ \<Rightarrow> m) F 0;
           suffix = replicate (Suc m) (CHR ''a'')
         in (\<lambda> x. Fun (Sharp (UnLab (x @ suffix))) [])"

abbreviation Tcap_Not_Unif :: "((string, 'l) lab, string, string) non_join_info"
where
  "Tcap_Not_Unif \<equiv> Tcap_Non_Unif default_grd_fun"

end
